﻿Shader "Custom/screen_space_curvature" 
{
	Properties 
	{		
		_MainTex ("Base (RGB)", 2D) = "yellow" {}
		_CurvTex ("Curvature Tex", 2D) = "red" {}		
	}
	SubShader 
	{
		Pass 
		{
			ZTest Always 
			Cull Off 
			ZWrite Off 
			//Fog { Mode Off } //Rendering settings

			//http://williamchyr.com/2013/11/unity-shaders-depth-and-normal-textures-part-3/
		
			CGPROGRAM			
			#pragma vertex vert_img			
			#pragma fragment frag

			//we include "UnityCG.cginc" to use the appdata_img struct
			#include "UnityCG.cginc"  


			#pragma target 4.0
			
			//sampler2D _CameraDepthNormalsTexture;
			sampler2D _CameraDepthTexture;

			sampler2D _CameraGBufferTexture2;
				
			sampler2D _MainTex;
			sampler2D _CurvTex;

			int _meanCurvature;
			int _useLod;
			float _inervalSize;

			float _fFarClipPlane;
			float3 _frustrumCorners[4];			
			float4x4 _CameraInverseViewMatrix;
			
			//==================================================================================
			//====================== Helper functions ==========================================
			//==================================================================================

			//http://docs.unity3d.com/462/Documentation/Manual/SL-BuiltinValues.html

			/**
			 * Get depth from texture
			 */
			float GetDepth(float2 vCenter)
			{				
				//https://github.com/keijiro/DeferredAO/blob/master/Assets/DeferredAO/Shader/DeferredAO.shader
				float depth = SAMPLE_DEPTH_TEXTURE(_CameraDepthTexture, vCenter);
		
				depth =  Linear01Depth(depth);
				//depth =  LinearEyeDepth (depth);
		
				return depth;
			}

			/**
			 * Map value s from interval [a1, a2] to interval [b1, b2]
			 */
			float MapRange(float a1, float a2, float b1, float b2, float s)
			{
				return b1 + (s - a1) * (b2 - b1) / (a2 - a1);
			}

			/**
			 * Reconstruct position in View Space
			 */
			float3 ReconstructPositionVS(float3 ray, float depth)
			{	
				return depth * ray;
			}

			/**
			 * Reconstruct position in World Space
			 */
			float3 ReconstructPositionWS(float3 ray, float depth)
			{	
				float3 pos = ReconstructPositionVS(ray, depth);

				pos = _WorldSpaceCameraPos + mul(_CameraInverseViewMatrix, float4(pos, 0)).xyz;  
	
				return pos;
			}

			/**
			 * Get ray from camera to the world in view space
			 */
			float3 GetRay(float2 pos)
			{
				float3 ray = lerp(lerp(_frustrumCorners[0], _frustrumCorners[1], pos.x),
		 					 lerp(_frustrumCorners[2], _frustrumCorners[3], pos.x),
							      pos.y);
		                 
				return ray;                 
			}



			/**
			 * Reconstruct position 
			 */
			float3 ReconstructPosition_textureCoord(float2 vCenter)
			{
				float3 ray = GetRay(vCenter);
	
				float depth = GetDepth(vCenter);

				float3 vPos = ReconstructPositionWS(ray, depth); 
     
				return vPos;
			}


			/**
			 * Reconstruct normal
			 */
			float3 ReconstructNormal_textureCoord(float2 vCenter)
			{	
				float3 worldSpaceNormal = tex2D( _CameraGBufferTexture2, vCenter).rgb * 2.0 - 1.0;
				return normalize(worldSpaceNormal);				
			}


			//=====================================================================================================================
			//============================= Curvature calculation =================================================================
			//=====================================================================================================================

			#define LOCAL_X  0
			#define LOCAL_Y  1
			#define LOCAL_Z  2
			#define TRIANGLE_VERTICES  3

			float2 CalcCurvature_ShapeOperator(float3 a, float3 b, float3 c, float3 an, float3 bn, float3 cn)
			{
	
				float3 local_trianglePos[TRIANGLE_VERTICES];
				float3 local_triangleNor[TRIANGLE_VERTICES];

	
				//=========================================================================================================

				float3 _ba_ = (b - a);
				float3 _ca_ = (c - a);

	
	
				float3 axis[TRIANGLE_VERTICES];
				axis[LOCAL_X] = normalize(_ba_);
				axis[LOCAL_Z] = normalize(cross(axis[LOCAL_X], _ca_));    
				axis[LOCAL_Y] = normalize(cross(axis[LOCAL_X], axis[LOCAL_Z]));
				
				float3x3 transformMatrix = float3x3(axis[LOCAL_X], axis[LOCAL_Y], axis[LOCAL_Z]);  //set rows
				
				local_trianglePos[0] = float3(0,0,0);
				local_triangleNor[0] = mul(transformMatrix, an);
	
				local_trianglePos[1] = mul(transformMatrix, _ba_);	
				local_triangleNor[1] = mul(transformMatrix, bn);

				local_trianglePos[2] = mul(transformMatrix, _ca_);	
				local_triangleNor[2] = mul(transformMatrix, cn);
	
				float3 u;
				float3 v;

				float3 du;
				float3 dv;

				u.x = local_trianglePos[1].x - local_trianglePos[0].x;
				u.y = local_trianglePos[2].x - local_trianglePos[1].x;
				u.z = local_trianglePos[0].x - local_trianglePos[2].x;

				v.x = local_trianglePos[1].y - local_trianglePos[0].y;
				v.y = local_trianglePos[2].y - local_trianglePos[1].y;
				v.z = local_trianglePos[0].y - local_trianglePos[2].y;
	

				du.x = local_triangleNor[1].x - local_triangleNor[0].x;
				du.y = local_triangleNor[2].x - local_triangleNor[1].x;
				du.z = local_triangleNor[0].x - local_triangleNor[2].x;

				dv.x = local_triangleNor[1].y - local_triangleNor[0].y;
				dv.y = local_triangleNor[2].y - local_triangleNor[1].y;	
				dv.z = local_triangleNor[0].y - local_triangleNor[2].y;
	
				//===============================================================================


				
				//inverse of 3x3 matrix
				//invB = inverse of B

				float aa = dot(u, u);
				float bb = dot(u, v);
				float cc = dot(v, v);

				float invB_B = -(bb * cc);
				float invB_C =  (bb * bb);
				float invB_D =  (aa * cc);
				float invB_E = -(aa * bb);
				float invB_A = -invB_C + (cc * cc) + invB_D;
				float invB_F =  (aa * aa) + invB_D - invB_C;

				float determinantB = 1.0 / ((aa + cc) * (invB_D - invB_C));



				//=======================================================================================

				//col based matrices
				float3x3 resLeft;
				float3x3 resRight;
	
				resLeft[0][0] = v.x * invB_B + u.x * invB_A;
				resLeft[1][0] = v.x * invB_C + u.x * invB_B;
				resLeft[2][0] = v.y * invB_B + u.y * invB_A;
				resRight[0][0] = v.y * invB_C + u.y * invB_B;
				resRight[1][0] = v.z * invB_B + u.z * invB_A;
				resRight[2][0] = v.z * invB_C + u.z * invB_B;

				resLeft[0][1] = v.x * invB_D + u.x * invB_B;
				resLeft[1][1] = v.x * invB_E + u.x * invB_D;
				resLeft[2][1] = v.y * invB_D + u.y * invB_B;
				resRight[0][1] = v.y * invB_E + u.y * invB_D;
				resRight[1][1] = v.z * invB_D + u.z * invB_B;
				resRight[2][1] = v.z * invB_E + u.z * invB_D;

				resLeft[0][2] = v.x * invB_E + u.x * invB_C;
				resLeft[1][2] = v.x * invB_F + u.x * invB_E;
				resLeft[2][2] = v.y * invB_E + u.y * invB_C;
				resRight[0][2] = v.y * invB_F + u.y * invB_E;
				resRight[1][2] = v.z * invB_E + u.z * invB_C;
				resRight[2][2] = v.z * invB_F + u.z * invB_E;

				//================================================================================
				
				float3 finalM_gpu_a = mul(float3(du.x, dv.x, du.y), resLeft);
				float3 finalM_gpu_b = mul(float3(dv.y, du.z, dv.z), resRight);

				float3 tmp_gpu = determinantB * (finalM_gpu_a + finalM_gpu_b);

				//================================================================================


				float D = dot(tmp_gpu, tmp_gpu) + (3.0 * tmp_gpu.y * tmp_gpu.y - 2.0 * tmp_gpu.x * tmp_gpu.z);
	
				if (D < 0) D = 0;
				float sqrtD = sqrt(D);

				float L1 = ((tmp_gpu.x + tmp_gpu.z) + sqrtD) * 0.5;
				float L2 = ((tmp_gpu.x + tmp_gpu.z) - sqrtD) * 0.5;


				float gauss = L1 * L2;
	
				float mean = 0.5 * (L1 + L2);

				return float2(gauss, mean);
			}




			float2 CalcCurvature_ShapeOperator_2edge(float3 a, float3 b, float3 c, float3 an, float3 bn, float3 cn)
			{
	
				float3 local_trianglePos[TRIANGLE_VERTICES];
				float3 local_triangleNor[TRIANGLE_VERTICES];

	
				//=========================================================================================================

				float3 _ba_ = (b - a);
				float3 _ca_ = (c - a);

	
	
				float3 axis[TRIANGLE_VERTICES];
				axis[LOCAL_X] = normalize(_ba_);
				axis[LOCAL_Z] = normalize(cross(axis[LOCAL_X], _ca_));    
				axis[LOCAL_Y] = normalize(cross(axis[LOCAL_X], axis[LOCAL_Z]));




				float3x3 transformMatrix = (float3x3(axis[LOCAL_X], axis[LOCAL_Y], axis[LOCAL_Z]));  //set columns
				

				local_trianglePos[0] = float3(0, 0, 0);
				local_triangleNor[0] = mul(transformMatrix, an);
	
				local_trianglePos[1] = mul(transformMatrix, _ba_);	
				local_triangleNor[1] = mul(transformMatrix, bn);

				local_trianglePos[2] = mul(transformMatrix, _ca_);	
				local_triangleNor[2] = mul(transformMatrix, cn);
	
				float2 u;
				float2 v;

				float2 du;
				float2 dv;

				u.x = local_trianglePos[1].x - local_trianglePos[0].x;
				u.y = local_trianglePos[2].x - local_trianglePos[1].x;
	
				v.x = local_trianglePos[1].y - local_trianglePos[0].y;
				v.y = local_trianglePos[2].y - local_trianglePos[1].y;
		
				du.x = local_triangleNor[1].x - local_triangleNor[0].x;
				du.y = local_triangleNor[2].x - local_triangleNor[1].x;
	
				dv.x = local_triangleNor[1].y - local_triangleNor[0].y;
				dv.y = local_triangleNor[2].y - local_triangleNor[1].y;	
		
				//===============================================================================
				

				//inverse of 3x3 matrix
				//invB = inverse of B

				float aa = dot(u, u);
				float bb = dot(u, v);
				float cc = dot(v, v);

				float invB_B = -(bb * cc);
				float invB_C =  (bb * bb);
				float invB_D =  (aa * cc);
				float invB_E = -(aa * bb);
				float invB_A = -invB_C + (cc * cc) + invB_D;
				float invB_F =  (aa * aa) + invB_D - invB_C;

				float determinantB = 1.0 / ((aa + cc) * (invB_D - invB_C));



				//=======================================================================================

				float4x4 tmp;
	
	
				tmp[0][0] = v.x * invB_B + u.x * invB_A;
				tmp[1][0] = v.x * invB_C + u.x * invB_B;
				tmp[2][0] = v.y * invB_B + u.y * invB_A;
				tmp[3][0] = v.y * invB_C + u.y * invB_B;


				tmp[0][1] = v.x * invB_D + u.x * invB_B;
				tmp[1][1] = v.x * invB_E + u.x * invB_D;
				tmp[2][1] = v.y * invB_D + u.y * invB_B;
				tmp[3][1] = v.y * invB_E + u.y * invB_D;


				tmp[0][2] = v.x * invB_E + u.x * invB_C;
				tmp[1][2] = v.x * invB_F + u.x * invB_E;
				tmp[2][2] = v.y * invB_E + u.y * invB_C;
				tmp[3][2] = v.y * invB_F + u.y * invB_E;


				tmp[0][3] = 0;
				tmp[1][3] = 0;
				tmp[2][3] = 0;
				tmp[3][3] = 0;


				//================================================================================

				float4 tmp_gpu = determinantB * mul(float4(du.x, dv.x, du.y, dv.y), tmp);
				
				//================================================================================


				float D = dot(tmp_gpu, tmp_gpu) + (3.0 * tmp_gpu.y * tmp_gpu.y - 2.0 * tmp_gpu.x * tmp_gpu.z);
	
				if (D < 0) D = 0;
				float sqrtD = sqrt(D);

				float L1 = ((tmp_gpu.x + tmp_gpu.z) + sqrtD) * 0.5;
				float L2 = ((tmp_gpu.x + tmp_gpu.z) - sqrtD) * 0.5;


				float gauss = L1 * L2;
	
				float mean = 0.5 * (L1 + L2);

				return float2(gauss, mean);
			}





			/**
			 * Calculate LOD
			 */
			float calcTextureLod(float2 texCoord)
			{
				// The OpenGL Graphics System: A Specification 4.2
				//  - chapter 3.9.11, equation 3.21
	
				float2 dx_vtc = ddx(texCoord);
				float2 dy_vtc = ddy(texCoord);
	
				float delta_max_sqr = max(dot(dx_vtc, dx_vtc), dot(dy_vtc, dy_vtc));

				return 0.5 * log2(delta_max_sqr); // == log2(sqrt(delta_max_sqr));		
			}

			int getSamplingStep(float2 texCoord, float depthCenter)
			{
				float3 posCenter = ReconstructPosition_textureCoord(texCoord);   
				float distCenter = length(posCenter);
				
				float2 tc = abs(posCenter.xz) / (_fFarClipPlane * depthCenter);
				float texLod = -calcTextureLod(tc);

				float f = (3 * texLod * (1 - depthCenter)) / log(distCenter);// / log(dd);

				return int(f);
			}


			//screen space triangulation neighbors
			//  TL TC TR
			//  LC C  RC
			//  BL BC BR

			#define DIR_TL 0
			#define DIR_TC 1
			#define DIR_TR 2

			#define DIR_LC 3
			#define DIR_C  4
			#define DIR_RC 5

			#define DIR_BL 6
			#define DIR_BC 7
			#define DIR_BR 8



			#define CURVATURE_FUNCTION CalcCurvature_ShapeOperator_2edge
			//#define CURVATURE_FUNCTION CalcCurvature_ShapeOperator

			/**
			 * Main fragment shader
			 */
			fixed4 frag (v2f_img input) : SV_Target 
			{
					
				float depthCenter = GetDepth(input.uv);
				
				//if (depthCenter < 0.0001)
				if (depthCenter > 0.9999)
				{
					return float4(0,0,0,1);	
				}  

				int2 dirs[9]; 
				int3 triangles[7];
				float3 pos[9];
				float3 nor[9];

				triangles[0] = int3(DIR_C, DIR_RC, DIR_TC);
				triangles[1] = int3(DIR_C, DIR_TC, DIR_TL);
				triangles[2] = int3(DIR_C, DIR_TL, DIR_LC);
				triangles[3] = int3(DIR_C, DIR_LC, DIR_BL);
				triangles[4] = int3(DIR_C, DIR_BL, DIR_BC);
				triangles[5] = int3(DIR_C, DIR_BC, DIR_BR);
				triangles[6] = int3(DIR_C, DIR_BR, DIR_RC);



				int iStep = 1;	
				if (_useLod == 1)
				{	
					iStep = getSamplingStep(input.uv, depthCenter);
				}
				
	
		 
				dirs[DIR_C] = int2(0, 0);

				dirs[DIR_TL] = int2(-iStep, iStep);
				dirs[DIR_TC] = int2(0, iStep);
				dirs[DIR_TR] = int2(iStep, iStep);

				dirs[DIR_LC] = int2(-iStep, 0);
				dirs[DIR_RC] = int2(iStep, 0);

				dirs[DIR_BL] = int2(-iStep , -iStep);
				dirs[DIR_BC] = int2(0, -iStep);
				dirs[DIR_BR] = int2(iStep , iStep);


				int2 center = input.uv * _ScreenParams.xy;

				for (int i = 0; i < 9; i++)
				{
					int2 c = center + dirs[i];

					float2 ct = float2(c.x + 0.5, c.y + 0.5) / _ScreenParams.xy; 

					
					float depth = GetDepth(ct);		
					
					if (depth < 0.0001)
					{
						c = center + sign(dirs[i]);
					}
					else if (abs(depthCenter - depth) >= 0.01)
					{
						c = center + sign(dirs[i]);
					}
					
		
					ct = float2(c.x + 0.5 , c.y + 0.5 ) / _ScreenParams.xy; 
					

					pos[i] = ReconstructPosition_textureCoord(ct);
					nor[i] = ReconstructNormal_textureCoord(ct);
				}								

				float2 curv = 0;

				if (iStep > 2)
				{		
					for (int i = 0; i < 7; i++)  
					{               
						curv += CURVATURE_FUNCTION(pos[triangles[i].x], pos[triangles[i].y], pos[triangles[i].z], 
									nor[triangles[i].x], nor[triangles[i].y], nor[triangles[i].z]);			
					}
					curv /= 7.0;
				}
				else 
				{
					curv += CURVATURE_FUNCTION(pos[DIR_C], pos[DIR_RC], pos[DIR_TC], 
									nor[DIR_C], nor[DIR_RC], nor[DIR_TC]);
				}


				float activeCurv = 0;
				if (_meanCurvature == 1) activeCurv = curv.y;
				else activeCurv = curv.x;

	
				

				float intervalSize = _inervalSize;
				activeCurv = clamp(activeCurv, -intervalSize, intervalSize);
	

				if (activeCurv < 0)
				{
					activeCurv = MapRange(-intervalSize, 0.0, 0.0, 0.49999, activeCurv);
				}
				else if (activeCurv > 0)
				{
					activeCurv = MapRange(0.0, intervalSize, 0.50001, 1.0, activeCurv);
				}
				else
				{
					activeCurv = 0.5;
				}
					
				

				float4 color = tex2D(_CurvTex, float2(activeCurv, 0.5));
				return color;			  				
			}
			ENDCG
		}
	} 
	FallBack off
}


