﻿using UnityEngine;
using System.Collections;
using UnityEngine.UI;

using System.Collections.Generic;

[ExecuteInEditMode]
public class CameraScript : MonoBehaviour 
{

	//========================================================================================
	// Public variables
	//========================================================================================

	public Texture CurvatureTexture;
	public Transform SceneRoot;
	public UnityEngine.UI.Text FrameTimeText;

	//========================================================================================
	// Private variables
	//========================================================================================


	private Material material = null;
	private ViewFrustum vf = null;
    private Vector3[] corners = new Vector3[4];

	private int meanCurvature = 1;
    private int useLOD = 1;
	private bool arcBallEnabled = false;

    private float intervalSize = 2.5f;

	private float lastTimeUpdate = 0;
	private int lastTimeUpdateCount = 0;

	private Vector2 actPoint;
	private Vector2 lastPoint;

	private GameObject loadedObject = null;

	private Dictionary<string, Transform> objectsInScene = new Dictionary<string, Transform>();

	//========================================================================================
	// Init
	//========================================================================================


	void Start()
	{
		Application.targetFrameRate = -1;

		//UnityEngine.Rendering.CullMode.Off;

		this.material = new Material(Shader.Find("Custom/screen_space_curvature"));
		this.material.SetTexture("_CurvTex", CurvatureTexture);

		//Camera.main.depthTextureMode = DepthTextureMode.DepthNormals;
		Camera.main.depthTextureMode = DepthTextureMode.Depth;

        this.intervalSize = 2.5f;
        this.useLOD = 0;
		this.arcBallEnabled = true;
        this.vf = new ViewFrustum(Camera.main);
        this.corners = new Vector3[4];


		for (int i = 0; i < this.SceneRoot.childCount; i++)
		{
			Transform t = this.SceneRoot.GetChild(i).transform;


			if (t.gameObject.name.StartsWith("Model"))
			{
				t.gameObject.SetActive(false);
				this.objectsInScene[t.gameObject.name] = t;

			}
		}


		this.objectsInScene["ModelDragon"].gameObject.SetActive(true);




		if (loadedObject == null)
		{
			string[] arguments = System.Environment.GetCommandLineArgs();

			if (arguments.Length != 1)
			{
				string fileName = arguments[1];
				//fileName = "D://mesh1.obj";

				Material mat = this.objectsInScene["ModelDragon"].GetComponentInChildren<MeshRenderer>().sharedMaterial;
				ModelLoader ml = new ModelLoader();
				loadedObject = ml.LoadModel(fileName, mat);

				loadedObject.transform.parent = SceneRoot;
				loadedObject.name = "ModelUserModel";

				loadedObject.SetActive(false);
				this.objectsInScene["ModelUserModel"] = loadedObject.transform;
			}
		}
				
		
		

		

	}

	void OnDestroy() 
	{
		foreach (var t in this.objectsInScene) 
		{
			if (t.Value != null)
			{
				t.Value.gameObject.SetActive(false);
			}
		}

		if (loadedObject == null) 
		{
			return;
		}

		loadedObject.transform.parent = null;
		Destroy(loadedObject);
		loadedObject = null;
	}

	//========================================================================================
	// GUI callbacks
	//========================================================================================

	public void ResetToDefault() 
	{
		this.SceneRoot.transform.position = Vector3.zero;
		this.SceneRoot.transform.rotation = Quaternion.identity;

		
	}

	public void ToogleModel(Toggle toggle) 
	{
		string name = toggle.name;
		name = name.Replace("Toggle", "Model");

		if (this.objectsInScene.ContainsKey(name)) 
		{
			bool act = this.objectsInScene[name].gameObject.activeSelf;
			this.objectsInScene[name].gameObject.SetActive(!act);
		}

	}


    public void ToggleLOD(Toggle toggle) 
    {
        if (toggle.isOn)
        {
            this.useLOD = 1;
        }
        else 
        {
            this.useLOD = 0;
        }
    }

	public void ToggleCurvature(Toggle tg) 
	{
		if (tg.isOn) 
		{
			this.meanCurvature = 1;
		}
		else 
		{
			this.meanCurvature = 0;
		}
	}

    public void ChangeIntervalSize(Slider slider) 
    {
        this.intervalSize = slider.value;
    }

    public void ChangeTextCaptionMax(Text text) 
    {
        text.text = "" + string.Format("{0:0.00}", this.intervalSize);
    }

	public void ChangeTextCaptionMin(Text text)
	{
		text.text = "-" + string.Format("{0:0.00}", this.intervalSize);
	}

	public void ToggleArcBall(Toggle toggle) 
	{
		this.arcBallEnabled = toggle.isOn;
	}



	//========================================================================================
	// ArcBall
	//========================================================================================


	void UpdatePicking()
	{
		//this->lastRay = this->actRay;
	
		
		this.lastPoint = this.actPoint;
	
		this.actPoint = new Vector2(Input.mousePosition.x, Input.mousePosition.y);

	

		//this->actRay  = this->camera->MapViewport2Dto3D(this->actPoint);	
	}


	Quaternion RotateLocalSpace(int selectedAxis, Vector3 pos, Quaternion rot)
	{
	

	

		Vector3 lastBall = this.MapToSphere(this.lastPoint);
		Vector3 actBall = this.MapToSphere(this.actPoint);

		float angle = Vector3.Angle(lastBall, actBall);

		Vector3 axis = new Vector3(0, 0, 0);
	
		if (selectedAxis != 0)
		{
		/*
			axis = SceneManagement::SceneEditorUtils::GetRotationAxis(selectedAxis);
	
			axis = MyMath::Vector3::TransformNormal(axis, rot->ToMatrix4x4());
			axis.Normalize();

			float dx, dy;
			this->CalcDxDy(pos, &dx, &dy);

			MyMath::Vector3 r = this->camera->GetRight();
			MyMath::Vector3 u = this->camera->GetUp();

			MyMath::Vector3 camDir = (r * dx + u * dy);
			camDir.Normalize();

			if (MyMath::Vector3::Dot(camDir, axis) < 0)
			{
				angle *= -1;
			}

		*/

		}
		else
		{
			axis = Vector3.Cross(actBall, lastBall);

			if (axis.sqrMagnitude < 0.0000001f)
			{
				return rot;
			}
	
			axis.Normalize();
		}




		Quaternion newRot = Quaternion.AngleAxis(angle, axis);
		newRot = newRot.Normalize();


	

		rot = rot * newRot;
	
		return rot;
	}
		

	Vector3 MapToSphere(Vector2 point) 	
	{

		//get current viewport settings
		float w = Camera.main.pixelWidth;// this->camera->GetViewport()->ClientWindowWidth;
		float h = Camera.main.pixelHeight;// this->camera->GetViewport()->ClientWindowHeight;


		
		//ball radius is 1, due to our mapping to <-1,1>, <-1, 1>
		float ballRadius = 1;
		float ballRadius2 = ballRadius * ballRadius;


		

		
		//map window from <0, width>, <0, height> to <-1, 1>, <-1, 1>
		//this put center of the arcball to the screen center [0, 0]

		float adjustedW = 1.0f / ((w - 1.0f) * 0.5f);
		float  adjustedH = 1.0f / ((h - 1.0f) * 0.5f);


		//convert scree point [x, y] to mapped screen point within <-1, 1>, <-1, 1>
		Vector2 mappedScreenPoint;
		mappedScreenPoint.x = (point.x * adjustedW) - 1.0f;
		mappedScreenPoint.y = 1.0f - (point.y * adjustedH); //Y coordinate is swapped
		
		//create final point
		Vector3 p = new Vector3(mappedScreenPoint.x, mappedScreenPoint.y, 0);
		
		

		//another approach without normalizing screen size to <-1,1>, <-1, 1>
		/*
		//http://www.raywenderlich.com/12667/how-to-rotate-a-3d-object-using-touches-with-opengl

		Vector3 ballCenter = new Vector3(w / 2, h / 2, 0);
		float ballRadius = Mathf.Min(w / 2, h / 2);
		float ballRadius2 = ballRadius * ballRadius;


		Vector3 screenPoint = new Vector3(point.x, point.y, 0);

		Vector3 p = screenPoint - ballCenter;		
		*/

		p.y *= -1;
	
		float pLength2 = p.sqrMagnitude;

		if (pLength2 <= ballRadius2)
		{
			//point is inside sphere
			//update Z coordinate
			p.z = Mathf.Sqrt(ballRadius2 - pLength2);
		}
		else
		{
			//point is outside sphere - map it to the closest point on the sphere
				
			//z-coordinate = 0
			p *= (ballRadius / p.magnitude);


			//p.Z = ballRadius2 / (2 * sqrt(pLength2));
			//float length = p.Length();
			//p /= length;

		}

		p.Normalize();

		return p;	
	}

	//========================================================================================
	// Updates
	//========================================================================================

	
	void Update() 
	{
		if (this.lastTimeUpdate > 0.1)
		{
			float t = this.lastTimeUpdate / this.lastTimeUpdateCount;

			this.FrameTimeText.text = "Frame time: " + string.Format("{0:0.00}", t * 1000.0) + "[ms]";
			this.lastTimeUpdate = 0;
			this.lastTimeUpdateCount = 0;
		}
		this.lastTimeUpdate += Time.deltaTime;
		this.lastTimeUpdateCount++;

		
		Vector3 pos = this.SceneRoot.transform.position;
		pos.z -= Input.mouseScrollDelta.y / 10;
		this.SceneRoot.transform.position = pos;

		this.UpdatePicking();

		if (this.arcBallEnabled == false) 
		{
			return;
		}

		if (Input.GetMouseButton(0))
		{
			//Ray ray = Camera.main.ScreenPointToRay(Input.mousePosition);
			//RaycastHit hit;

			


			Quaternion q = this.RotateLocalSpace(0, SceneRoot.position, SceneRoot.rotation);

			SceneRoot.rotation = q;

		}

		
	}


	void OnRenderImage(RenderTexture source, RenderTexture destination)
	{
        

        this.vf.GetCorners(ref this.corners, ViewFrustum.Face.BACK, ViewFrustum.Space.VIEW);

        //set array
        //http://answers.unity3d.com/questions/979080/how-to-pass-an-array-to-a-shader.html
        material.SetVector("_frustrumCorners0", this.corners[0]);
        material.SetVector("_frustrumCorners1", this.corners[1]);
        material.SetVector("_frustrumCorners2", this.corners[2]);
        material.SetVector("_frustrumCorners3", this.corners[3]);

        material.SetInt("_useLod", this.useLOD);
		material.SetInt("_meanCurvature", this.meanCurvature);

        //Matrix4x4 inv = Matrix4x4.Inverse(Camera.main.projectionMatrix);
        //material.SetMatrix("_CameraInverseProjectionMatrix", inv);

        Matrix4x4 inv = Matrix4x4.Inverse(Camera.main.worldToCameraMatrix);
		material.SetMatrix("_CameraInverseViewMatrix", inv);

        material.SetFloat("_fFarClipPlane", Camera.main.farClipPlane);

        material.SetFloat("_inervalSize", this.intervalSize);


		
			

        //sphere.transform.position = this.corners[2];

		Graphics.Blit(source, destination, material);
	}
}
